// Project Sword & Gun Copyright © 2021 Kevin Poretti


#include "Characters/SNGCharacterBase.h"

#include "DrawDebugHelpers.h"
#include "Camera/CameraComponent.h"
#include "Components/CapsuleComponent.h"
#include "Components/SNGCharacterMovementComponent.h"
#include "GameFramework/SpringArmComponent.h"
#include "Kismet/KismetMathLibrary.h"


static int32 ShowCharacterMovementDebug = 0;
FAutoConsoleVariableRef CVARShowCharacterMovementDebug(TEXT("SNG.ShowCharacterMovementDebug"),
						ShowCharacterMovementDebug, 
						TEXT("Print movement information for character"), 
						ECVF_Cheat);

// Sets default values
ASNGCharacterBase::ASNGCharacterBase(const FObjectInitializer& ObjectInitializer)
	: Super(ObjectInitializer.SetDefaultSubobjectClass<USNGCharacterMovementComponent>(ACharacter::CharacterMovementComponentName))
{
 	// Set this character to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;

	SpringArmComponent = CreateDefaultSubobject<USpringArmComponent>(TEXT("SpringArmComponent"));
	SpringArmComponent->SetupAttachment(RootComponent);
	SpringArmComponent->bUsePawnControlRotation = true;
	SpringArmComponent->bInheritYaw = true;
	SpringArmComponent->bInheritPitch = true;
	SpringArmComponent->bInheritRoll = true;
	
	CameraComponent = CreateDefaultSubobject<UCameraComponent>(TEXT("CameraComponent"));
	CameraComponent->SetupAttachment(SpringArmComponent);
	
	bUseControllerRotationYaw = false;
	bUseControllerRotationPitch = false;
	bUseControllerRotationRoll = false;
	
	USkeletalMeshComponent* CharMesh = GetMesh();
	if(CharMesh)
	{
		CharMesh->SetRelativeLocation(FVector(0.0f, 0.0f, -GetCapsuleComponent()->GetScaledCapsuleHalfHeight()));
		CharMesh->SetRelativeRotation(FRotator(0.0f, -90.0f, 0.0f));
	}
	else
	{
		UE_LOG(LogTemp, Warning, TEXT("SNGCharacterBase :: Could not get skeletal mesh. Did you forget to set it?"));
	}
}

void ASNGCharacterBase::MoveForward(float Value)
{
	if(!CanMove)
	{
		return;
	}
	
	InputDirection.Y = Value;

	UAnimMontage* CurrMontage = GetCurrentMontage();
	if(CurrMontage && !InputDirection.IsNearlyZero(0.1) && !GetCharacterMovement()->IsFalling() && CanMovementCancelMontage)
	{
		StopAnimMontage(CurrMontage);
	}
	
	FRotator ViewRotation = FRotator(0.0f, GetControlRotation().Yaw, 0.0f);
	AddMovementInput(UKismetMathLibrary::GetForwardVector(ViewRotation), Value);
}

void ASNGCharacterBase::MoveRight(float Value)
{
	if(!CanMove)
	{
		return;
	}
	
	InputDirection.X = Value;

	UAnimMontage* CurrMontage = GetCurrentMontage();
	if(CurrMontage && !InputDirection.IsNearlyZero(0.1) && !GetCharacterMovement()->IsFalling() && CanMovementCancelMontage)
	{
		StopAnimMontage(CurrMontage);
	}
	
	FRotator ViewRotation = FRotator(0.0f, GetControlRotation().Yaw, 0.0f);
	AddMovementInput(UKismetMathLibrary::GetRightVector(ViewRotation), Value);
}

void ASNGCharacterBase::StartJump()
{
	if(!CanJump())
	{
		return;
	}

	// can't rotate in air
	GetCharacterMovement()->bOrientRotationToMovement = false;

	CanMovementCancelMontage = false;
	
	FVector ActorDir = GetActorForwardVector();
	FVector InputDir = GetInputAsWorldDirection();
	float Dot = UKismetMathLibrary::Dot_VectorVector(ActorDir, InputDir);
	if(Dot > 0.5f)
	{
		check(JumpFwdMontage);
		PlayAnimMontage(JumpFwdMontage);
	}
	else if (Dot < -0.5f)
	{
		check(JumpBackMontage);
		PlayAnimMontage(JumpBackMontage);
	}
	else
	{
		check(JumpNeutralMontage);
		PlayAnimMontage(JumpNeutralMontage);
	}
}

void ASNGCharacterBase::Landed(const FHitResult& Hit)
{
	Super::Landed(Hit);

	// can rotate when on the ground again
	GetCharacterMovement()->bOrientRotationToMovement = true;
	CanMovementCancelMontage = true;
	
	PlayAnimMontage(JumpNeutralMontage, 1.0f, "Land");
}

// Called when the game starts or when spawned
void ASNGCharacterBase::BeginPlay()
{
	Super::BeginPlay();

	CanMovementCancelMontage = false;
	CanMove = true;
}

FVector ASNGCharacterBase::GetInputAsWorldDirection()
{
	AController* MyController = GetController();
	if(MyController == nullptr)
	{
		return FVector::ZeroVector;
	}
	
	FRotator ViewRotation = FRotator(0.0f, MyController->GetControlRotation().Yaw, 0.0f);
	
	FVector ForwardInput = UKismetMathLibrary::GetForwardVector(ViewRotation) * InputDirection.Y;

	FVector RightInput = UKismetMathLibrary::GetRightVector(ViewRotation) * InputDirection.X;

	return ForwardInput + RightInput;
}

// Called every frame
void ASNGCharacterBase::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);

	if(ShowCharacterMovementDebug)
	{
		FVector StartPos = GetActorLocation() - GetCapsuleComponent()->GetScaledCapsuleHalfHeight();
		FVector ActorForward = UKismetMathLibrary::GetForwardVector(GetActorRotation());
		DrawDebugDirectionalArrow(GetWorld(), StartPos, StartPos + (ActorForward * 100.0f),
			10.0f, FColor::Cyan, false, -1.0f, 0, 2.0f);
		
		FVector CurrInputDirection = GetInputAsWorldDirection();
		CurrInputDirection.Normalize();
		DrawDebugDirectionalArrow(GetWorld(), StartPos, StartPos + (CurrInputDirection * 100.0f),
		10.0f, FColor::Magenta, false, -1.0f, 0, 2.0f);
	}
	
}

// Called to bind functionality to input
void ASNGCharacterBase::SetupPlayerInputComponent(UInputComponent* PlayerInputComponent)
{
	Super::SetupPlayerInputComponent(PlayerInputComponent);

	// char movement
    PlayerInputComponent->BindAxis("MoveForward", this, &ASNGCharacterBase::MoveForward);
	PlayerInputComponent->BindAxis("MoveRight", this, &ASNGCharacterBase::MoveRight);

	// camera movement
	PlayerInputComponent->BindAxis("Turn", this, &ASNGCharacterBase::AddControllerYawInput);
	PlayerInputComponent->BindAxis("LookUp", this, &ASNGCharacterBase::AddControllerPitchInput);

	PlayerInputComponent->BindAction("Jump", IE_Pressed, this, &ASNGCharacterBase::StartJump);
	PlayerInputComponent->BindAction("Jump", IE_Released, this, &ASNGCharacterBase::StopJumping);
}

